/* N-body calculation using Greenspan 1st-order discrete mechanics.	*/
/* 1-D version with vector calculation, mixed spreading and systolic	*/
/* computation.  All processors store particle information and		*/
/* calculate velocities, positions redundantly.				*/
/* Uses dynamic shape allocation to permit more flexible operation.	*/
/* This versions uses O(1) storage per vp, rather than O(NS) as in	*/
/* nb1dsys.cs, calculating the Greenspan accelerations from scratch.	*/
/* File:  nbigsys.cs							*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <cscomm.h>
#include <float.h>
#include "nbigsys.hs"

/* Spread xv[3] along i-axis (0-axis) from diagonal, leaving result in dv, */
/* Called with all processors active and leaves all active.  Uses wv.	   */

void spreadi(float:pair *dv, float:pair *xv)
 { int u;
   for ( u = 0;  u < 3;  u++ )
    { where (*diag) wv[u] = xv[u]; else wv[u] = -FLT_MAX;
      dv[u] = spread(wv[u], 0, CMC_combiner_max);
    }
 }

/* Sum accelerations, leaving results in all processors of each row.	*/
/* Called with all processors active.					*/

void sumac()
 { /* Return moving ac's to original positions */
  everywhere
  {to_torus_dim(acm, acm, vsz, 2, SS);
   /* print_x("acm after return", acm, 2); 	For testing */
   /* Accumulate moving ac's along i axis */
   acm[0] = spread(acm[0], 0, CMC_combiner_add);
   acm[1] = spread(acm[1], 0, CMC_combiner_add);
   acm[2] = spread(acm[2], 0, CMC_combiner_add);
   /* Combine on diagonal */
   where (*diag) { ac[0] += acm[0];  ac[1] += acm[1];  ac[2] += acm[2]; }
   /* Sum accelerations along j-axis */
   ac[0] = spread(ac[0], 1, CMC_combiner_add);
   ac[1] = spread(ac[1], 1, CMC_combiner_add);
   ac[2] = spread(ac[2], 1, CMC_combiner_add);
   /* print_x("ac after accumulation", ac, 2);	For testing */
  } /* end everywhere */
 }


/* Calculate accelerations.  x0j, mj already have "0" positions and	*/
/* scaled masses for "j" particles.  Called with all processors active.	*/

void accel()
 { float:pair d0[3], d1[3], r0, r1, w, f;  int u, s;
  everywhere
  {spreadi(xm1, x1);		      /* Spread x1 along i-axis into xm1 */
   /* Now xm1 = [j][j][k]x1,  x1 = [i][i][k]x1 */
   /* print_x("accel - xm1 after spreadi", xm1, 2);	/* For testing */
   *mmj = *mj;				/* Load conveyer */
   r0 = r1 = 0.0f;			/* To accumulate sum of squares */
   for ( u = 0;  u < 3;  u++ )	      /* First acceleration calculation */
    { ac[u] = acm[u] = 0.0f;   xm0[u] = x0j[u];	       /* Load conveyor */
      d0[u] = xm0[u] - x0[u];  r0 += d0[u]*d0[u];
      d1[u] = xm1[u] - x1[u];  r1 += d1[u]*d1[u]; }
   r0 = sqrt(r0);  r1 = sqrt(r1);		/* Euclidean distances */
   /* print_x("accel - r1", &r1, 2);	/* For testing */
   w = *mmj/(r0*r1*(r0 + r1));			/* Multiplier */
   where (*act)					/* Form acceleration */
    { for ( u = 0;  u < 3;  u++ ) ac[u] = w*(d0[u] + d1[u]); }
   /* print_x("accel - ac after step 0", ac, 2);	For testing */
   /* Loop for intermediate (systolic) calculations */
   for ( s = 1;  s < SS;  s++)
    { to_torus_dim(xm0, xm0, cvsz, 2, 1);		/* Advance conveyor */
      r0 = r1 = 0.0f;
      for ( u = 0;  u < 3;  u++ )
       { d0[u] = xm0[u] - x0[u];  r0 += d0[u]*d0[u];
	 d1[u] = xm1[u] - x1[u];  r1 += d1[u]*d1[u];
         wv[u] = d0[u] + d1[u]; }
      r0 = sqrt(r0);  r1 = sqrt(r1);		/* Euclidean distances */
      w = r0*r1*(r0 + r1);
      f = *mi/w;  w = *mmj/w; 	 	/* Multipliers */
      for ( u = 0;  u < 3;  u++ )		/* Form accelerations */
       { ac[u] += w*wv[u];  acm[u] -= f*wv[u]; }
      /* print_x("accel - ac after step", ac, 2);	For testing */
      /* print_x("accel - acm after step", acm, 2);	For testing */
    } /* End systolic loop */
   to_torus_dim(xm0, xm0, cvsz, 2, 1);		/* Advance conveyor */
   r0 = r1 = 0.0f;
   for ( u = 0;  u < 3;  u++ )
    { d0[u] = xm0[u] - x0[u];  r0 += d0[u]*d0[u];
      d1[u] = xm1[u] - x1[u];  r1 += d1[u]*d1[u];
      wv[u] = d0[u] + d1[u]; }
   r0 = sqrt(r0);  r1 = sqrt(r1);		/* Euclidean distances */
   w = *mmj/(r0*r1*(r0 + r1));			/* Multiplier */
   for ( u = 0;  u < 3;  u++ ) ac[u] += w*wv[u];	/* Accelerations */
   /* print_x("accel - final ac", ac, 2);		For testing */
   /* print_x("accel - final acm", acm, 2);		For testing */
   sumac();					/* Add them up */
  } /* end everywhere */
 } /* end accel */
